@@ -17,16 +17,16 @@ namespace camp {
 
 const triple drawElement::zero;
 
+//double Tx[3]; // x-component of current transform
+//double Ty[3]; // y-component of current transform
+double Tz[3]; // z-component of current transform
+
 using vm::array;
 
 #ifdef HAVE_GL
-void bezierTriangle(const triple *g, bool straight, double ratio,
-                    const triple& Min, const triple& Max,
-                    GLfloat *colors);
-  
-void bezierPatch(const triple *g, bool straight, double ratio,
-                 const triple& Min, const triple& Max,
-                 GLfloat *colors);
+BezierCurve drawSurface::C;
+BezierPatch drawBezierPatch::S;
+BezierTriangle drawBezierTriangle::S;
 
 void storecolor(GLfloat *colors, int i, const vm::array &pens, int j)
 {
@@ -52,8 +52,30 @@ void setcolors(bool colors, bool lighton,
                const RGBAColour& emissive,
                const RGBAColour& specular, double shininess) 
 {
+  static prc::RGBAColour lastdiffuse;
+  static prc::RGBAColour lastambient;
+  static prc::RGBAColour lastemissive;
+  static prc::RGBAColour lastspecular;
+  static double lastshininess=0.0;
+  static bool lastcolors=false;
+    
+  if(colors != lastcolors) {
+    drawBezierPatch::S.draw();
+    lastcolors=colors;
+  }
+  
+  if(!colors && (diffuse != lastdiffuse || ambient != lastambient || 
+                 emissive != lastemissive || specular != lastspecular ||
+                 shininess != lastshininess)) {
+    drawBezierPatch::S.draw(); 
+    lastdiffuse=diffuse;
+    lastambient=ambient;
+    lastemissive=emissive;
+    lastspecular=specular;
+    lastshininess=shininess;
+  }
+
   if(colors) {
-    glEnable(GL_COLOR_MATERIAL);
     if(!lighton) 
       glColorMaterial(GL_FRONT_AND_BACK,GL_EMISSION);
 
@@ -86,13 +108,13 @@ void setcolors(bool colors, bool lighton,
 
 #endif  
 
-void drawSurface::bounds(const double* t, bbox3& b)
+void drawBezierPatch::bounds(const double* t, bbox3& b)
 {
   double x,y,z;
   double X,Y,Z;
   
   if(straight) {
-    triple Vertices[3];
+    triple Vertices[4];
     if(t == NULL) {
       Vertices[0]=controls[0];
       Vertices[1]=controls[3];
@@ -152,7 +174,7 @@ void drawSurface::bounds(const double* t, bbox3& b)
   }
 }
 
-void drawSurface::ratio(const double* t, pair &b, double (*m)(double, double),
+void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double),
                         double fuzz, bool &first)
 {
   triple buf[16];
@@ -206,7 +228,7 @@ void drawSurface::ratio(const double* t, pair &b, double (*m)(double, double),
   }
 }
 
-bool drawSurface::write(prcfile *out, unsigned int *, double, groupsmap&)
+bool drawBezierPatch::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible || !prc)
     return true;
@@ -225,9 +247,10 @@ bool drawSurface::write(prcfile *out, unsigned int *, double, groupsmap&)
   return true;
 }
 
-void drawSurface::render(GLUnurbs *nurb, double size2,
-                         const triple& b, const triple& B,
-                         double perspective, bool lighton, bool transparent)
+void drawBezierPatch::render(GLUnurbs *nurb, double size2,
+                             const triple& b, const triple& B,
+                             double perspective, bool lighton,
+                             bool transparent)
 {
 #ifdef HAVE_GL
   if(invisible || 
@@ -257,12 +280,26 @@ void drawSurface::render(GLUnurbs *nurb, double size2,
   glGetDoublev(GL_MODELVIEW_MATRIX,t);
 // Like Fortran, OpenGL uses transposed (column-major) format!
   run::transpose(t,4);
+/*  
+  Tx[0]=t[0];
+  Tx[1]=t[1];
+  Tx[2]=t[2];
+  
+  Ty[0]=t[4];
+  Ty[1]=t[5];
+  Ty[2]=t[6];
+*/
+  
+  Tz[0]=t[8];
+  Tz[1]=t[9];
+  Tz[2]=t[10];
+  
   run::inverse(t,4);
   bbox3 box(m,M);
   box.transform(t);
   m=box.Min();
   M=box.Max();
-
+  
   if(!billboard && (Max.getx() < m.getx() || Min.getx() > M.getx() ||
                     Max.gety() < m.gety() || Min.gety() > M.gety() ||
                     Max.getz() < m.getz() || Min.getz() > M.getz()))
@@ -278,9 +315,9 @@ void drawSurface::render(GLUnurbs *nurb, double size2,
       storecolor(v,4*i,colors[i]);
   
   triple *Controls;
-  triple C[16];
+  triple Controls0[16];
   if(billboard) {
-    Controls=C;
+    Controls=Controls0;
     for(size_t i=0; i < 16; i++)
       Controls[i]=BB.transform(controls[i]);
   } else
@@ -288,25 +325,26 @@ void drawSurface::render(GLUnurbs *nurb, double size2,
     
   if(gl::outlinemode) {
     triple edge0[]={Controls[0],Controls[4],Controls[8],Controls[12]};
-    R.render(edge0,straight,size3.length()/size2,m,M);
+    C.queue(edge0,straight,size3.length()/size2,m,M);
     triple edge1[]={Controls[12],Controls[13],Controls[14],Controls[15]};
-    R.render(edge1,straight,size3.length()/size2,m,M);
+    C.queue(edge1,straight,size3.length()/size2,m,M);
     triple edge2[]={Controls[15],Controls[11],Controls[7],Controls[3]};
-    R.render(edge2,straight,size3.length()/size2,m,M);
+    C.queue(edge2,straight,size3.length()/size2,m,M);
     triple edge3[]={Controls[3],Controls[2],Controls[1],Controls[0]};
-    R.render(edge3,straight,size3.length()/size2,m,M);
-    R.draw();
-  } else
-    bezierPatch(Controls,straight,size3.length()/size2,m,M,colors ? v : NULL);
-
-  if(colors)
-    glDisable(GL_COLOR_MATERIAL);
+    C.queue(edge3,straight,size3.length()/size2,m,M);
+    C.draw();
+  } else {
+    S.queue(Controls,straight,size3.length()/size2,m,M,transparent,
+            colors ? v : NULL);
+    if(!transparent) 
+      S.draw();
+  }
 #endif
 }
 
-drawElement *drawSurface::transformed(const double* t)
+drawElement *drawBezierPatch::transformed(const double* t)
 {
-  return new drawSurface(t,this);
+  return new drawBezierPatch(t,this);
 }
   
 void drawBezierTriangle::bounds(const double* t, bbox3& b)
@@ -480,6 +518,20 @@ void drawBezierTriangle::render(GLUnurbs *nurb, double size2,
   glGetDoublev(GL_MODELVIEW_MATRIX,t);
 // Like Fortran, OpenGL uses transposed (column-major) format!
   run::transpose(t,4);
+/*  
+  Tx[0]=t[0];
+  Tx[1]=t[1];
+  Tx[2]=t[2];
+  
+  Ty[0]=t[4];
+  Ty[1]=t[5];
+  Ty[2]=t[6];
+*/
+  
+  Tz[0]=t[8];
+  Tz[1]=t[9];
+  Tz[2]=t[10];
+
   run::inverse(t,4);
   bbox3 box(m,M);
   box.transform(t);
@@ -501,9 +553,9 @@ void drawBezierTriangle::render(GLUnurbs *nurb, double size2,
       storecolor(v,4*i,colors[i]);
     
   triple *Controls;
-  triple C[10];
+  triple Controls0[10];
   if(billboard) {
-    Controls=C;
+    Controls=Controls0;
     for(size_t i=0; i < 10; i++)
       Controls[i]=BB.transform(controls[i]);
   } else 
@@ -511,17 +563,18 @@ void drawBezierTriangle::render(GLUnurbs *nurb, double size2,
   
   if(gl::outlinemode) {
     triple edge0[]={Controls[0],Controls[1],Controls[3],Controls[6]};
-    R.render(edge0,straight,size3.length()/size2,m,M);
+    C.queue(edge0,straight,size3.length()/size2,m,M);
     triple edge1[]={Controls[6],Controls[7],Controls[8],Controls[9]};
-    R.render(edge1,straight,size3.length()/size2,m,M);
+    C.queue(edge1,straight,size3.length()/size2,m,M);
     triple edge2[]={Controls[9],Controls[5],Controls[2],Controls[0]};
-    R.render(edge2,straight,size3.length()/size2,m,M);
-    R.draw();
-  } else
-    bezierTriangle(Controls,straight,size3.length()/size2,m,M,colors ? v : NULL);
-  
-  if(colors)
-    glDisable(GL_COLOR_MATERIAL);
+    C.queue(edge2,straight,size3.length()/size2,m,M);
+    C.draw();
+  } else {
+    S.queue(Controls,straight,size3.length()/size2,m,M,transparent,
+            colors ? v : NULL);
+    if(!transparent) 
+      S.draw();
+  }
 #endif
 }
 
