@@ -14,6 +14,7 @@
 #include "drawverbatim.h"
 #include "drawlabel.h"
 #include "drawlayer.h"
+#include "drawsurface.h"
 
 using std::ifstream;
 using std::ofstream;
@@ -32,10 +33,13 @@ texstream::~texstream() {
   unlink((name+"aux").c_str());
   unlink((name+"log").c_str());
   unlink((name+"out").c_str());
-  unlink((name+"m9").c_str());
-  if(settings::pdf(texengine))
+  if(settings::pdf(texengine)) {
     unlink((name+"pdf").c_str());
+    unlink((name+"m9").c_str());
+  } else
+    unlink((name+"pbsdat").c_str());
   if(context) {
+    unlink("cont-new.log");
     unlink((name+"tex").c_str());
     unlink((name+"top").c_str());
     unlink((name+"tua").c_str());
@@ -45,6 +49,8 @@ texstream::~texstream() {
 
 namespace camp {
 
+extern void draw();
+
 bool isIdTransform3(const double* t)
 {
   return (t == NULL || (t[0]==1 && t[1]==0 && t[2]==0 && t[3]==0 &&
@@ -232,6 +238,26 @@ bool picture::have3D()
   return false;
 }
 
+bool picture::havepng()
+{
+  for(nodelist::iterator p=nodes.begin(); p != nodes.end(); ++p) {
+    assert(*p);
+    if((*p)->svgpng())
+      return true;
+  }
+  return false;
+}
+
+bool picture::havenewpage()
+{
+  for(nodelist::iterator p=nodes.begin(); p != nodes.end(); ++p) {
+    assert(*p);
+    if((*p)->isnewpage())
+      return true;
+  }
+  return false;
+}
+
 bbox picture::bounds()
 {
   size_t n=nodes.size();
@@ -403,7 +429,7 @@ int opentex(const string& texname, const string& prefix, bool dvi)
 
 bool picture::texprocess(const string& texname, const string& outname,
                          const string& prefix, const pair& bboxshift,
-                         bool svgformat)
+                         bool svg)
 {
   int status=1;
   ifstream outfile;
@@ -421,10 +447,10 @@ bool picture::texprocess(const string& texname, const string& outname,
       string dviname=auxname(prefix,"dvi");
       mem::vector<string> cmd;
     
-      if(svgformat) {
+      if(svg) {
         cmd.push_back(getSetting<string>("dvisvgm"));
         cmd.push_back("-n");
-        cmd.push_back("--verbosity=3");
+        cmd.push_back("-v0");
         string libgs=getSetting<string>("libgs");
         if(!libgs.empty())
           cmd.push_back("--libgs="+libgs);
@@ -469,10 +495,14 @@ bool picture::texprocess(const string& texname, const string& outname,
           cmd.push_back("-Pdownload35");
           cmd.push_back("-D600");
           cmd.push_back("-O"+String(hoffset)+"bp,"+String(voffset)+"bp");
-          cmd.push_back("-T"+String(getSetting<double>("paperwidth"))+"bp,"+
-                        String(paperHeight)+"bp");
+          bool ps=havenewpage();
+          if(ps)
+            cmd.push_back("-T"+String(getSetting<double>("paperwidth"))+"bp,"+
+                          String(paperHeight)+"bp");
+          else
+            cmd.push_back("-E");
           push_split(cmd,getSetting<string>("dvipsOptions"));
-          if(getSetting<string>("papertype") != "")
+          if(ps && getSetting<string>("papertype") != "")
             cmd.push_back("-t"+papertype);
           if(verbose <= 1) cmd.push_back("-q");
           cmd.push_back("-o"+psname);
@@ -498,7 +528,7 @@ bool picture::texprocess(const string& texname, const string& outname,
                   continue;
 
                 if(s.find("%!PS-Adobe-") == 0) {
-                  fout.header();
+                  fout.header(!ps);
                   continue;
                 }
 
@@ -596,6 +626,35 @@ int picture::epstopdf(const string& epsname, const string& pdfname)
   return status;
 }
   
+int picture::pdftoeps(const string& pdfname, const string& epsname)
+{
+  mem::vector<string> cmd;
+  cmd.push_back(getSetting<string>("gs"));
+  cmd.push_back("-q");
+  cmd.push_back("-dNOCACHE");
+  cmd.push_back("-dNOPAUSE");
+  cmd.push_back("-dBATCH");
+  cmd.push_back("-P");
+  if(safe)
+    cmd.push_back("-dSAFER");
+  string texengine=getSetting<string>("tex");
+  cmd.push_back("-sDEVICE="+getSetting<string>("epsdriver"));
+  
+  cmd.push_back("-sOutputFile="+stripDir(epsname));
+  cmd.push_back(stripDir(pdfname));
+
+  char *oldPath=NULL;
+  string dir=stripFile(epsname);
+  if(!dir.empty()) {
+    oldPath=getPath();
+    setPath(dir.c_str());
+  }
+  int status=System(cmd,0,true,"gs","Ghostscript");
+  if(oldPath != NULL)
+    setPath(oldPath);
+  return status;
+}
+  
 bool picture::reloadPDF(const string& Viewer, const string& outname) const 
 {
   static bool needReload=true;
@@ -611,7 +670,7 @@ bool picture::reloadPDF(const string& Viewer, const string& outname) const
     needReload=false;
     string texengine=getSetting<string>("tex");
     Setting("tex")=string("pdflatex");
-    haveReload=f.shipout(NULL,reloadprefix,"pdf",0.0,false,false);
+    haveReload=f.shipout(NULL,reloadprefix,"pdf",false,false);
     Setting("tex")=texengine;
   }
   if(haveReload) {
@@ -626,17 +685,55 @@ bool picture::reloadPDF(const string& Viewer, const string& outname) const
   return true;
 }               
   
-  
+int picture::epstosvg(const string& epsname, const string& outname)
+{  
+  mem::vector<string> cmd;
+  cmd.push_back(getSetting<string>("dvisvgm"));
+  cmd.push_back("-n");
+  cmd.push_back("-E");
+  cmd.push_back("--verbosity=3");
+  string libgs=getSetting<string>("libgs");
+  if(!libgs.empty())
+    cmd.push_back("--libgs="+libgs);
+  push_split(cmd,getSetting<string>("dvisvgmOptions"));
+  cmd.push_back("-o"+outname);
+  cmd.push_back(epsname);
+  int status=System(cmd,2,true,"dvisvgm");
+  if(!getSetting<bool>("keep"))
+    unlink(epsname.c_str());
+  return status;
+}
+
+int picture::pdftosvg(const string& pdfname, const string& outname)
+{  
+  mem::vector<string> cmd;
+  cmd.push_back(getSetting<string>("dvisvgm"));
+  cmd.push_back("-n");
+  cmd.push_back("--pdf");
+  cmd.push_back("--verbosity=3");
+  string libgs=getSetting<string>("libgs");
+  if(!libgs.empty())
+    cmd.push_back("--libgs="+libgs);
+  push_split(cmd,getSetting<string>("dvisvgmOptions"));
+  cmd.push_back("-o"+outname);
+  cmd.push_back(pdfname);
+  int status=System(cmd,2,true,"dvisvgm");
+  if(status == 0 && !getSetting<bool>("keep"))
+    unlink(pdfname.c_str());
+  return status;
+}
+
 bool picture::postprocess(const string& prename, const string& outname,
-                          const string& outputformat, double magnification,
-                          bool wait, bool view, bool pdftex, bool svgformat)
+                          const string& outputformat,
+                          bool wait, bool view, bool pdftex, 
+                          bool epsformat, bool svg)
 {
   static mem::map<CONST string,int> pids;
   int status=0;
-  bool epsformat=outputformat == "eps";
   bool pdfformat=(settings::pdf(getSetting<string>("tex")) 
                   && outputformat == "") || outputformat == "pdf";
   
+  mem::vector<string> cmd;
   if(pdftex || !epsformat) {
     if(pdfformat) {
       if(pdftex) {
@@ -644,8 +741,19 @@ bool picture::postprocess(const string& prename, const string& outname,
         if(status != 0)
           reportError("Cannot rename "+prename+" to "+outname);
       } else status=epstopdf(prename,outname);
+    } else if(epsformat) {
+      if(svg) {
+        status=pdftosvg(prename,outname);
+        if(status != 0) { // Dvisvgm version < 2.4 doesn't support --pdf
+          string epsname=stripExt(prename)+".eps";
+          status=pdftoeps(prename,epsname);
+          if(status != 0) return false;
+          status=epstosvg(epsname,outname);
+        }
+        epsformat=false;
+      } else 
+        status=pdftoeps(prename,outname);
     } else {
-      mem::vector<string> cmd;
       double render=fabs(getSetting<double>("render"));
       if(render == 0) render=1.0;
       double res=render*72.0;
@@ -665,7 +773,7 @@ bool picture::postprocess(const string& prename, const string& outname,
         cmd.push_back("-sOutputFile="+outname);
         cmd.push_back(prename);
         status=System(cmd,0,true,"gs","Ghostscript");
-      } else if(!svgformat) {
+      } else if(!svg && !getSetting<bool>("xasy")) {
         double expand=antialias;
         if(expand < 2.0) expand=1.0;
         res *= expand;
@@ -675,9 +783,10 @@ bool picture::postprocess(const string& prename, const string& outname,
         if(expand == 1.0)
           cmd.push_back("+antialias");
         push_split(cmd,getSetting<string>("convertOptions"));
-        cmd.push_back("-geometry");
+        cmd.push_back("-resize");
         cmd.push_back(String(100.0/expand)+"%x");
-        cmd.push_back(nativeformat()+":"+prename);
+        if(outputformat == "jpg") cmd.push_back("-flatten");
+        cmd.push_back(prename);
         cmd.push_back(outputformat+":"+outname);
         status=System(cmd,0,true,"convert");
       }
@@ -754,8 +863,7 @@ string Outname(const string& prefix, const string& outputformat,
 }
 
 bool picture::shipout(picture *preamble, const string& Prefix,
-                      const string& format, double magnification,
-                      bool wait, bool view)
+                      const string& format, bool wait, bool view)
 {
   b=bounds();
   
@@ -771,28 +879,38 @@ bool picture::shipout(picture *preamble, const string& Prefix,
   string outputformat=format.empty() ? defaultformat() : format;
   bool epsformat=outputformat == "eps";
   bool pdfformat=pdf || outputformat == "pdf";
-  bool svgformat=outputformat == "svg" && !pdf && usetex &&
+  bool svgformat=outputformat == "svg";
+  bool dvi=false;
+  bool svg=svgformat && usetex && !(pdf && havepng()) &&
     (!have3D() || getSetting<double>("render") == 0.0);
+  if(svg) {
+    if(pdf) epsformat=true;
+    else dvi=true;
+  }
   
-  bool xobject=magnification > 0;
   string outname=Outname(prefix,outputformat,standardout);
   string epsname=epsformat ? (standardout ? "" : outname) :
     auxname(prefix,"eps");
   
   bool Labels=labels || TeXmode;
   
-  if(b.empty && !Labels) { // Output a null file
+  bool empty=b.empty;
+  if(outputformat == "png" && (b.right-b.left < 1.0 || b.top-b.bottom < 1.0))
+    empty=true;
+
+  if(empty && !Labels) { // Output a null file
     bbox b;
     b.left=b.bottom=0;
-    b.right=b.top=xobject ? 18 : 1;
+    b.right=b.top=1;
     psfile out(epsname,false);
     out.prologue(b);
     out.epilogue();
     out.close();
-    return postprocess(epsname,outname,outputformat,1.0,wait,view,false,false);
+    return postprocess(epsname,outname,outputformat,wait,view,false,
+                       epsformat,false);
   }
   
-  Labels |= svgformat;
+  Labels |= svg;
     
   if(Labels)
     prefix=cleanpath(prefix);
@@ -800,13 +918,6 @@ bool picture::shipout(picture *preamble, const string& Prefix,
   string prename=((epsformat && !pdf) || !Labels) ? epsname : 
     auxname(prefix,preformat);
   
-  if(xobject) {
-    double fuzz=0.5/magnification;
-    b.top += fuzz;
-    b.right += fuzz;
-    b.bottom -= fuzz;
-  }
-    
   SetPageDimensions();
   
   pair aligndir=getSetting<pair>("aligndir");
@@ -841,7 +952,7 @@ bool picture::shipout(picture *preamble, const string& Prefix,
   
   if(Labels) {
     texname=TeXmode ? buildname(prefix,"tex") : auxname(prefix,"tex");
-    tex=svgformat ? new svgtexfile(texname,b) : new texfile(texname,b);
+    tex=dvi ? new svgtexfile(texname,b) : new texfile(texname,b);
     tex->prologue();
   }
   
@@ -896,7 +1007,7 @@ bool picture::shipout(picture *preamble, const string& Prefix,
     bool postscript=false;
     drawLabel *L=NULL;
     
-    if(svgformat)
+    if(dvi)
       for(nodelist::const_iterator r=begin.begin(); r != begin.end(); ++r)
         (*r)->draw(&out);
     
@@ -904,7 +1015,7 @@ bool picture::shipout(picture *preamble, const string& Prefix,
       assert(*p);
       if(Labels && (*p)->islayer()) break;
       
-      if(svgformat && (*p)->svg()) {
+      if(dvi && (*p)->svg()) {
         picture *f=(*p)->svgpng() ? new picture : NULL;
         nodelist::const_iterator q=layerp;
         for(;;) {
@@ -950,7 +1061,7 @@ bool picture::shipout(picture *preamble, const string& Prefix,
           buf << prefix << "_" << svgcount;
           ++svgcount;
           string pngname=buildname(buf.str(),"png");
-          f->shipout(preamble,buf.str(),"png",0.0,false,false);
+          f->shipout(preamble,buf.str(),"png",false,false);
           pair m=f->bounds().Min();
           pair M=f->bounds().Max();
           delete f;
@@ -991,7 +1102,7 @@ bool picture::shipout(picture *preamble, const string& Prefix,
         for (p=layerp; p != nodes.end(); ++p) {
           assert(*p);
           bool islayer=(*p)->islayer();
-          if(svgformat && (*p)->svg()) {
+          if(dvi && (*p)->svg()) {
             islayer=true;
             if((*p)->svgpng())
               L->write(tex,b);
@@ -1019,8 +1130,8 @@ bool picture::shipout(picture *preamble, const string& Prefix,
       if(Labels) {
         tex->epilogue();
         if(context) prefix=stripDir(prefix);
-        status=texprocess(texname,svgformat ? outname : prename,prefix,
-                          bboxshift,svgformat);
+        status=texprocess(texname,dvi ? outname : prename,prefix,
+                          bboxshift,dvi);
         delete tex;
         if(!getSetting<bool>("keep")) {
           for(mem::list<string>::iterator p=files.begin(); p != files.end();
@@ -1029,15 +1140,12 @@ bool picture::shipout(picture *preamble, const string& Prefix,
         }
       }
       if(status) {
-        if(xobject) {
-          if(pdf || transparency)
-            status=(epstopdf(prename,Outname(prefix,"pdf",standardout)) == 0);
-        } else {
-          if(context) prename=stripDir(prename);
-          status=postprocess(prename,outname,outputformat,magnification,wait,
-                             view,pdf && Labels,svgformat);
-          if(pdfformat && !getSetting<bool>("keep"))
-            unlink(auxname(prefix,"m9").c_str());
+        if(context) prename=stripDir(prename);
+        status=postprocess(prename,outname,outputformat,wait,
+                           view,pdf && Labels,epsformat,svg);
+        if(pdfformat && !getSetting<bool>("keep")) {
+          unlink(auxname(prefix,"m9").c_str());
+          unlink(auxname(prefix,"pbsdat").c_str());
         }
       }
     }
@@ -1057,6 +1165,9 @@ void picture::render(GLUnurbs *nurb, double size2,
     assert(*p);
     (*p)->render(nurb,size2,Min,Max,perspective,lighton,transparent);
   }
+#ifdef HAVE_GL
+  drawBezierPatch::S.draw();
+#endif  
 }
   
 struct Communicate : public gc {
